import glob

# 1: Create Data from s3.
# python -m lightning create data --name example --source s3://pl-flash-data/wiki-test/

# 2: Attach the Data to an experiment
# python -m lightning run experiment data.py --data example

for filename in glob.iglob("/data/example/**/**", recursive=True):
    print(filename)



import urllib.parse
from typing import List

from lightning.app.api import Post
from lightning.app.storage import Drive

from lightning_hpo.commands.tensorboard.stop import StopTensorboardConfig, TensorboardConfig
from lightning_hpo.components.tensorboard import Tensorboard
from lightning_hpo.controllers.controller import Controller
from lightning_hpo.utilities.enum import Stage


class TensorboardController(Controller):

    model = TensorboardConfig

    def on_reconcile_start(self, configs: List[TensorboardConfig]):
        for config in configs:
            work_name = urllib.parse.quote_plus(config.sweep_id)
            if work_name not in self.r:
                if config.stage in (Stage.STOPPED, Stage.NOT_STARTED) and config.desired_stage == Stage.RUNNING:
                    self.r[work_name] = Tensorboard(
                        drive=Drive(f"lit://{config.sweep_id}"),
                        config=config,
                    )
                    self.r[work_name].stage = Stage.PENDING

    def show_tensorboards(self) -> List[TensorboardConfig]:
        """Show TensorBoards."""
        if self.db_url:
            return self.db.select_all()
        return []

    def run_tensorboard(self, config: TensorboardConfig):
        """Run TensorBoard for a given Sweep or Experiment."""
        tensorboards = self.db.select_all()
        matched_tensorboard = None

        for tensorboard in tensorboards:
            if tensorboard.sweep_id == config.sweep_id:
                matched_tensorboard = config

        if matched_tensorboard:
            matched_tensorboard.stage = Stage.STOPPED
            matched_tensorboard.desired_stage = Stage.RUNNING
            self.db.update(matched_tensorboard)
            return f"Re-Launched a Tensorboard `{config.sweep_id}`."

        self.db.insert(config)
        return f"Launched a Tensorboard `{config.sweep_id}`."

    def stop_tensorboard(self, config: StopTensorboardConfig):
        """Stop TensorBoard for a given Sweep or Experiment."""
        work_name = urllib.parse.quote_plus(config.sweep_id)
        if work_name in self.r:
            self.r[work_name].stop()
            self.r[work_name]._url = ""
            self.r[work_name].stage = Stage.STOPPED
            self.r[work_name].desired_stage = Stage.STOPPED
            self.db.update(self.r[work_name].collect_model())
            del self.r[work_name]
            return f"Tensorboard `{config.sweep_id}` was stopped."
        return f"Tensorboard `{config.sweep_id}` doesn't exist."

    def configure_api(self):
        return [
            Post("/api/run/tensorboard", self.run_tensorboard),
            Post("/api/stop/tensorboard", self.stop_tensorboard),
            Post("/api/show/tensorboards", self.show_tensorboards),
        ]



class Stage:
    NOT_STARTED = "not_started"
    PENDING = "pending"
    RUNNING = "running"
    PRUNED = "pruned"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    STOPPED = "stopped"
    STOPPING = "stopping"



__version__ = "0.0.4"
__author__ = "PyTorchLightning et al."
__author_email__ = "thomas@grid.ai"
__license__ = "TBD"
__copyright__ = f"Copyright (c) 2021-2022, {__author__}."
__homepage__ = "https://github.com/PyTorchLightning/lightning"
__docs__ = "Lightning HPO"
__long_doc__ = """
Lightning HPO provides a pythonic implementation for Scalable Hyperparameter Tuning.
This library relies on Optuna for providing state-of-the-art sampling hyper-parameters algorithms and efficient experiment pruning strategies.
"""

__all__ = [
    "__author__",
    "__author_email__",
    "__copyright__",
    "__docs__",
    "__homepage__",
    "__license__",
    "__version__",
]



import os
from argparse import ArgumentParser
from getpass import getuser
from typing import List
from uuid import uuid4

from lightning.app.utilities.commands import ClientCommand
from sqlalchemy import Column
from sqlmodel import Field, SQLModel

from lightning_hpo.utilities.enum import Stage
from lightning_hpo.utilities.utils import pydantic_column_type


class NotebookConfig(SQLModel, table=True):

    __table_args__ = {"extend_existing": True}

    notebook_name: str = Field(primary_key=True)
    requirements: List[str] = Field(..., sa_column=Column(pydantic_column_type(List[str])))
    cloud_compute: str
    stage: str = Stage.NOT_STARTED
    desired_stage: str = Stage.RUNNING
    url: str = ""
    start_time: float = -1


class RunNotebookCommand(ClientCommand):
    def run(self) -> None:
        parser = ArgumentParser()

        parser.add_argument("--name", default=None, help="The name of your notebook to run.")
        parser.add_argument(
            "--requirements",
            default=[],
            type=lambda s: [v.replace(" ", "") for v in s.split(",")] if "," in s else s,
            help="List of requirements separated by a comma or requirements.txt filepath.",
        )
        parser.add_argument(
            "--cloud_compute",
            default="cpu",
            choices=["cpu", "cpu-small", "cpu-medium", "gpu", "gpu-fast", "gpu-fast-multi"],
            type=str,
            help="The machine to use in the cloud.",
        )

        hparams, _ = parser.parse_known_args()

        if isinstance(hparams.requirements, str) and os.path.exists(hparams.requirements):
            with open(hparams.requirements, "r") as f:
                hparams.requirements = [line.replace("\n", "") for line in f.readlines()]

        id = str(uuid4()).split("-")[0]
        notebook_name = hparams.name or f"{getuser()}-{id}"

        config = NotebookConfig(
            notebook_name=notebook_name,
            requirements=hparams.requirements,
            cloud_compute=hparams.cloud_compute,
        )
        response = self.invoke_handler(config=config)
        print(response)



from abc import ABC, abstractmethod
from typing import Any, Dict, List

from lightning_hpo.commands.sweep.run import ExperimentConfig


class Algorithm(ABC):
    @abstractmethod
    def register_distributions(self, distributions):
        ...

    @abstractmethod
    def register_experiments(self, experiments: List[ExperimentConfig]):
        ...

    @abstractmethod
    def experiment_start(self, experiment_id: int):
        ...

    @abstractmethod
    def experiment_end(self, experiment_id: int, score: float):
        ...

    @abstractmethod
    def should_prune(self, experiment_id: int, reports: List[float]) -> bool:
        ...

    @abstractmethod
    def get_params(self, experiment_id: int) -> Dict[str, Any]:
        ...



from argparse import ArgumentParser

from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel


class StopExperimentConfig(BaseModel):
    name: str


class StopExperimentCommand(ClientCommand):

    description = "Stop an Experiment."

    def run(self) -> None:
        parser = ArgumentParser()
        parser.add_argument("--name", type=str, required=True, help="The associated `experiment_name` to stop.")
        hparams = parser.parse_args()
        response = self.invoke_handler(config=StopExperimentConfig(name=hparams.name))
        print(response)



import os

import pytorch_lightning as pl
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torchmetrics.functional import accuracy
from torchvision import transforms
from torchvision.datasets import CIFAR10


class LitModel(pl.LightningModule):
    def __init__(self, lr: float = 0.0001, batch_size: int = 32):
        super().__init__()
        self.save_hyperparameters()
        self.layer_1 = torch.nn.Linear(3 * 32 * 32, 128)
        self.layer_2 = torch.nn.Linear(128, 10)

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = self.layer_1(x)
        x = F.relu(x)
        x = self.layer_2(x)
        x = F.log_softmax(x)
        return x

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.lr)
        return optimizer

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = F.nll_loss(y_hat, y)
        self.log("train_loss", loss)
        self.log("train_acc", accuracy(y_hat.exp(), y), prog_bar=True)
        return loss


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("--gpus", type=int, default=None)
    parser.add_argument("--lr", type=float, default=1e-3)
    parser.add_argument("--batch_size", type=int, default=32)
    parser.add_argument("--max_epochs", type=int, default=10)
    parser.add_argument("--data_dir", type=str, default=os.getcwd())
    args = parser.parse_args()

    dataset = CIFAR10(args.data_dir, download=True, transform=transforms.ToTensor())
    train_loader = DataLoader(dataset, batch_size=args.batch_size)

    # init model
    model = LitModel(lr=args.lr)

    # most basic trainer, uses good defaults (auto-tensorboard, checkpoints, logs, and more)
    trainer = pl.Trainer(gpus=args.gpus, max_epochs=args.max_epochs)
    trainer.fit(model, train_loader)



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, TypedDict


class DistributionDict(TypedDict):
    distribution: str
    params: Dict[str, Any]


class Distribution(ABC):
    @abstractmethod
    def to_dict(self) -> DistributionDict:
        ...


class LogUniform(Distribution):
    def __init__(self, low: float, high: float) -> None:
        self.low = low
        self.high = high

    def to_dict(self) -> DistributionDict:
        return {"distribution": "log_uniform", "params": {"low": self.low, "high": self.high}}


class Uniform(Distribution):
    def __init__(self, low: float, high: float) -> None:
        self.low = low
        self.high = high

    def to_dict(self):
        return {"distribution": "uniform", "params": {"low": self.low, "high": self.high}}


class Categorical(Distribution):
    def __init__(self, choices: List[Any]) -> None:
        self.choices = choices

    def to_dict(self) -> DistributionDict:
        return {"distribution": "categorical", "params": {"choices": self.choices}}


class IntUniform(Distribution):
    def __init__(self, low: int, high: int, step: Optional[int] = 1) -> None:
        self.low = low
        self.high = high
        self.step = step

    def to_dict(self) -> DistributionDict:
        return {"distribution": "int_uniform", "params": {"low": self.low, "high": self.high, "step": self.step}}



from enum import Enum

from lightning_hpo.loggers.logger import Logger, NoneLogger
from lightning_hpo.loggers.streamlit.streamlit import StreamLitLogger
from lightning_hpo.loggers.tensorboard import TensorboardLogger
from lightning_hpo.loggers.wandb import WandbLogger


class LoggerType(Enum):
    STREAMLIT = "streamlit"
    WANDB = "wandb"
    TENSORBOARD = "tensorboard"
    NONE = "none"

    def get_logger(self) -> Logger:
        if self == LoggerType.NONE:
            return NoneLogger()
        if self == LoggerType.STREAMLIT:
            return StreamLitLogger()
        elif self == LoggerType.WANDB:
            return WandbLogger()
        elif self == LoggerType.TENSORBOARD:
            return TensorboardLogger()
        else:
            raise ValueError("Unknown runtime type")



from abc import abstractmethod
from typing import List, Optional, Type

from lightning import LightningFlow
from lightning.app.components.database import DatabaseClient
from lightning.app.storage import Drive
from lightning.app.structures import Dict
from sqlmodel import SQLModel

from lightning_hpo.utilities.enum import Stage
from lightning_hpo.utilities.utils import get_primary_key


class ControllerResource:

    model: Type[SQLModel]

    def on_collect_model(self, model_dict):
        """Override to add the missing elements to the model_dict."""

    def collect_model(self):
        keys = list(self.model.__fields__)
        model_dict = {key: getattr(self, key) for key in keys if key in self._state}
        self.on_collect_model(model_dict)
        return self.model.parse_obj(model_dict)


class Controller(LightningFlow):

    model: Type[SQLModel]

    def __init__(self, drive: Optional[Drive] = None):
        super().__init__()
        self.db_url = None
        self._token = None
        self.r = Dict()
        self.drive = drive
        self._db_client = None
        self.ready = False

    def run(self, db_url: str, token: str, configs: Optional[List[Type[SQLModel]]] = None):
        self.db_url = db_url
        self._token = token

        # TODO: Resolve scheduling. It seems only the last one is activated somehow in the cloud.
        # TODO: Improve the schedule API.
        # 1: Read from the database and generate the works accordingly.
        # if self.schedule("* * * * * 0,5,10,15,20,25,30,35,40,45,50,55"):
        db_configs = self.db.select_all()
        if not self.ready:
            for config in db_configs:
                config.stage = Stage.NOT_STARTED
                self.db.update(config)
            self.ready = True

        if configs:
            db_configs += db_configs
        if db_configs:
            self.on_reconcile_start(db_configs)

        # 2: Iterate over the.r and collect updates
        configs = []
        for resource in self.r.values():
            resource.run()
            configs.append(resource.collect_model())

        if not configs:
            return

        # 3: Reconciler on end
        primary_key = get_primary_key(self.model)
        db_configs = {getattr(config, primary_key): config for config in db_configs}
        for config in configs:
            db_config = db_configs[getattr(config, primary_key)]
            if config.dict() != db_config.dict():
                self.db.update(config)

        self.on_reconcile_end(configs)

    @property
    def db(self) -> DatabaseClient:
        if self._db_client is None:
            assert self.db_url is not None
            self._db_client = DatabaseClient(self.db_url, self._token, model=self.model)
        return self._db_client

    @abstractmethod
    def on_reconcile_start(self, configs):
        """Override with your start reconciliation mechanism."""

    @abstractmethod
    def on_reconcile_end(self, updates):
        """Override with your end reconciliation mechanism."""



from argparse import ArgumentParser

from lightning.app.utilities.commands import ClientCommand
from sqlmodel import SQLModel


class DeleteDataConfig(SQLModel):
    name: str


class DeleteDataCommand(ClientCommand):

    description = "Delete Data."

    def run(self) -> None:
        parser = ArgumentParser()

        parser.add_argument("--name", type=str, required=True, help="The name of the Data.")

        hparams = parser.parse_args()
        response = self.invoke_handler(config=DeleteDataConfig(name=hparams.name))
        print(response)



from typing import Optional

from sqlmodel import Field, SQLModel

from lightning_hpo.utilities.enum import Stage


class TensorboardConfig(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    sweep_id: str
    shared_folder: str
    stage: str = Stage.NOT_STARTED
    desired_stage: str = Stage.RUNNING
    url: str = ""


class StopTensorboardConfig(SQLModel):
    sweep_id: str



import argparse
from typing import Any, Dict, List, Optional

import requests
from lightning.app.utilities.commands import ClientCommand
from pydantic import BaseModel
from rich.console import Console
from rich.table import Table

from lightning_hpo.commands.sweep.run import SweepConfig


def _show_empty_sweep():
    table = Table(
        "id",
        "cloud_compute",
        "total_experiments",
        "total_experiments_done",
        title="Sweep",
        show_header=True,
        header_style="bold green",
    )

    console = Console()
    console.print(table)


def _show_sweeps(sweeps: List[SweepConfig]):
    if not sweeps:
        _show_empty_sweep()

    for sweep in sweeps:
        _show_sweep(sweep)


def _parse_params(params):
    out = {}
    for k, v in params.items():
        if isinstance(v, float) and v == int(v):
            out[k] = int(v)
        else:
            out[k] = v
    return out


def _show_sweep(sweep: SweepConfig):
    table = Table(
        "id",
        "cloud_compute",
        "total_experiments",
        "total_experiments_done",
        title="Sweep",
        show_header=True,
        header_style="bold green",
    )

    table.add_row(
        sweep.sweep_id,
        sweep.cloud_compute,
        str(sweep.total_experiments),
        str(sweep.total_experiments_done),
    )
    console = Console()
    console.print(table)

    params = list(sweep.experiments[0].params)
    monitor = sweep.experiments[0].monitor

    table = Table(
        "name",
        "progress",
        "best_model_score",
        *params,
        title=f"Experiments monitor=({monitor})",
        show_header=True,
        header_style="bold green",
    )

    for experiment in sweep.experiments.values():
        table.add_row(
            str(experiment.name),
            str(experiment.progress) if experiment.stage != "failed" else "failed",
            str(round(experiment.best_model_score, 2) if experiment.best_model_score else None),
            *[str(round(v, 5)) for v in _parse_params(experiment.params).values()],
        )
    console = Console()
    console.print(table)


class ShowSweepsCommand(ClientCommand):

    description = "Show all Sweeps or the Experiments from a given Sweep."

    # TODO: (tchaton) Upstream to Lightning
    def invoke_handler(self, config: Optional[BaseModel] = None) -> Dict[str, Any]:
        command = self.command_name.replace(" ", "_")
        resp = requests.post(self.app_url + f"/command/{command}", data=config.json() if config else None)
        assert resp.status_code == 200, resp.json()
        return resp.json()

    def run(self) -> None:
        # 1. Parse the user arguments.
        parser = argparse.ArgumentParser()
        parser.add_argument("--name", type=str, default=None, help="Provide the `name` to be showed.")
        hparams = parser.parse_args()

        # 2: Collect the SweepConfig
        resp = self.invoke_handler()

        # 3: Display the Sweeps or Sweep
        # TODO: Undestand why the format isn't the same
        try:
            sweeps = [SweepConfig.parse_raw(sweep) for sweep in resp]
        except Exception:
            sweeps = [SweepConfig(**sweep) for sweep in resp]

        if hparams.name is None:
            _show_sweeps(sweeps)
        else:
            matching_sweep = [sweep for sweep in sweeps if sweep.sweep_id == hparams.name]
            if not matching_sweep:
                raise Exception(
                    "The provided `sweep_id` doesn't exists. "
                    f"Here are the available ones {[s.sweep_id for s in sweeps]}"
                )
            assert len(matching_sweep) == 1
            _show_sweep(*matching_sweep)



import argparse
import math
import os
from pathlib import Path
from typing import List, Optional, Tuple

import requests
from lightning.app.utilities.commands import ClientCommand
from lightning.app.utilities.network import LightningClient
from lightning_app.utilities.cloud import _get_project
from pydantic import BaseModel
from tqdm.auto import tqdm

from lightning_hpo.commands.artifacts.show import _collect_artifact_paths, _filter_paths, ShowArtifactsConfig


def _download_file(url: str, target_file: str):
    target_file = Path(target_file).resolve()
    target_file.parent.mkdir(exist_ok=True, parents=True)

    response = requests.get(url, allow_redirects=True, stream=True)
    total_length = response.headers.get("content-length", None)

    with open(target_file, "wb") as target:
        if total_length is None:
            target.write(response.content)
        else:
            total_length = int(total_length)
            chunk_size = 1024  # 1 KB
            for chunk in tqdm(
                response.iter_content(chunk_size=chunk_size),
                total=int(total_length / chunk_size),
                unit="KB",
                desc=str(target_file.name),
                leave=True,  # progressbar stays
            ):
                target.write(chunk)


def _copy_file(source_file: str, target_file: str):
    source_file = Path(source_file).resolve()
    target_file = Path(target_file).resolve()
    target_file.parent.mkdir(exist_ok=True, parents=True)

    total_length = os.stat(source_file).st_size

    with open(target_file, "wb") as target:
        with open(source_file, "rb") as source:
            chunk_size = 1024  # 1 KB
            chunks = math.ceil(total_length / chunk_size)
            for i in tqdm(
                range(chunks),
                total=chunks,
                unit="KB",
                desc=str(target_file.name),
                leave=True,  # progressbar stays
            ):
                target.write(source.read(chunk_size))


class DownloadArtifactsConfig(BaseModel):
    include: Optional[str] = None
    exclude: Optional[str] = None


class DownloadArtifactsConfigResponse(BaseModel):
    sweep_names: List[str]
    experiment_names: List[str]
    paths: List[str]
    urls: Optional[List[str]]


class DownloadArtifactsCommand(ClientCommand):

    description = "Download an artifact."

    def run(self) -> None:
        # 1. Parse the user arguments.
        parser = argparse.ArgumentParser()
        parser.add_argument("--names", nargs="+", default=[], help="Provide a list of experiments or sweep names.")
        parser.add_argument(
            "--output_dir", required=True, type=str, help="Provide the output directory for the artifacts.."
        )
        hparams = parser.parse_args()

        output_dir = Path(hparams.output_dir).resolve()

        if not output_dir.exists():
            output_dir.mkdir(exist_ok=True)

        config = DownloadArtifactsConfig(include=None, exclude=None)
        response = DownloadArtifactsConfigResponse(**self.invoke_handler(config=config))

        for name in hparams.names:
            has_found = False
            if name in response.sweep_names:
                has_found = True
            elif name in response.experiment_names:
                has_found = True

            if not has_found:
                raise ValueError(
                    f"The provided name `{name}` doesn't exists. "
                    f" Here are the available Sweeps {response.sweep_names} and "
                    f"available Experiments {response.experiment_names}."
                )

        if response.urls:
            for path, url in zip(response.paths, response.urls):
                if not any(name in path for name in hparams.names):
                    continue
                if path.startswith("/"):
                    path = path[1:]
                _download_file(url, os.path.join(output_dir, str(path).split("drive/")[1]))
        else:
            for path in response.paths:
                if not any(name in path for name in hparams.names):
                    continue
                _copy_file(path, os.path.join(output_dir, str(path).split("drive/")[1]))

        print("All the specified artifacts were downloaded.")


def _collect_artifact_urls(config: DownloadArtifactsConfig) -> Tuple[List[str], Optional[List[str]]]:
    """This function is responsible to collect the files from the shared filesystem."""

    use_localhost = "LIGHTNING_APP_STATE_URL" not in os.environ

    if use_localhost:
        paths = _collect_artifact_paths(
            config=ShowArtifactsConfig(include=config.include, exclude=config.exclude),
            replace=False,
        )
        return paths, None
    else:
        client = LightningClient()
        project_id = _get_project(client).project_id
        app_id = os.getenv("LIGHTNING_CLOUD_APP_ID")
        response = client.lightningapp_instance_service_list_lightningapp_instance_artifacts(project_id, app_id)
        paths = [artifact.filename for artifact in response.artifacts]
        maps = {artifact.filename: artifact.url for artifact in response.artifacts}
        filtered_paths = _filter_paths(paths, include=config.include, exclude=config.exclude)
        return filtered_paths, [maps[path] for path in filtered_paths]



import os.path as ops

from lightning import LightningApp

from lightning_hpo import Sweep
from lightning_hpo.distributions.distributions import Uniform

app = LightningApp(
    Sweep(
        script_path=ops.join(ops.dirname(__file__), "scripts/objective.py"),
        total_experiments=3,
        parallel_experiments=1,
        direction="maximize",
        distributions={"x": Uniform(-10, 10)},
    )
)



from lightning.app.utilities.commands import ClientCommand
from rich.console import Console
from rich.table import Table


class ShowDataCommand(ClientCommand):

    description = "List all Data."

    def run(self) -> None:
        response = self.invoke_handler()

        table = Table(
            "name",
            "source",
            "mount_path",
            title="Data",
            show_header=True,
            header_style="bold green",
        )

        for mount in response:
            table.add_row(
                mount["name"],
                mount["source"],
                mount["mount_path"],
            )
        console = Console()
        console.print(table)



import os.path as ops

from lightning import LightningApp

from lightning_hpo import Sweep
from lightning_hpo.algorithm import GridSearch

app = LightningApp(
    Sweep(
        script_path=ops.join(ops.dirname(__file__), "scripts/objective.py"),
        total_experiments=3,
        parallel_experiments=1,
        algorithm=GridSearch(search_space={"x": [-10, 0, 10]}),
    )
)



"""Root package info."""
import logging
import os

_root_logger = logging.getLogger()
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.INFO)

_console = logging.StreamHandler()
_console.setLevel(logging.INFO)

formatter = logging.Formatter("%(levelname)s: %(message)s")
_console.setFormatter(formatter)

# if root logger has handlers, propagate messages up and let root logger process them,
# otherwise use our own handler
if not _root_logger.hasHandlers():
    _logger.addHandler(_console)
    _logger.propagate = False

from lightning_hpo.__about__ import *  # noqa: E402, F401, F403
from lightning_hpo.components.sweep import Sweep  # noqa: E402
from lightning_hpo.framework.agnostic import Objective  # noqa: E402
from lightning_hpo.utilities.utils import HPOCloudCompute  # noqa: E402, F401

_PACKAGE_ROOT = os.path.dirname(__file__)
_PROJECT_ROOT = os.path.dirname(_PACKAGE_ROOT)

from optuna.storages._in_memory import _logger

_logger.disabled = True
_logger.propagate = False

__all__ = ["Objective", "Sweep", "HPOCloudCompute"]



import time
from typing import Any, Dict, Optional

from lightning.app.components.training import LightningTrainingComponent, PyTorchLightningScriptRunner

from lightning_hpo.framework.agnostic import Objective


class PyTorchLightningObjective(Objective, PyTorchLightningScriptRunner):

    """This component executes a PyTorch Lightning script
    and injects a callback in the Trainer at runtime in order to start tensorboard server."""

    def __init__(
        self,
        *args,
        logger: str,
        sweep_id: str,
        experiment_id: int,
        experiment_name: str,
        num_nodes: int,
        **kwargs,
    ):
        Objective.__init__(
            self,
            logger=logger,
            sweep_id=sweep_id,
            experiment_id=experiment_id,
            experiment_name=experiment_name,
            **kwargs,
        )
        PyTorchLightningScriptRunner.__init__(self, *args, num_nodes=num_nodes, **kwargs)
        self.progress = None
        self.total_parameters = None
        self.start_time = None
        self.end_time = None

    def configure_tracer(self):
        tracer = Objective.configure_tracer(self)
        if self.node_rank == 0:
            return self.add_metadata_tracker(tracer)
        return tracer

    def run(self, params: Optional[Dict[str, Any]] = None, restart_count: int = 0, **kwargs):
        self.params = params
        return PyTorchLightningScriptRunner.run(self, params=params, **kwargs)

    def on_after_run(self, script_globals):
        self.end_time = time.time()
        PyTorchLightningScriptRunner.on_after_run(self, script_globals)
        self.best_model_path = str(self.best_model_path)

    @classmethod
    def distributions(cls):
        return None

    def add_metadata_tracker(self, tracer):
        import pytorch_lightning as pl
        from pytorch_lightning.callbacks import Callback, DeviceStatsMonitor
        from pytorch_lightning.strategies.deepspeed import DeepSpeedStrategy
        from pytorch_lightning.utilities import rank_zero_only
        from pytorch_lightning.utilities.model_summary.model_summary import (
            _is_lazy_weight_tensor,
            get_human_readable_count,
        )
        from pytorch_lightning.utilities.model_summary.model_summary_deepspeed import deepspeed_param_size

        class ProgressCallback(Callback):
            def __init__(self, work):
                self.work = work
                self.work.start_time = time.time()
                self.device_stats_callback = DeviceStatsMonitor(cpu_stats=True)

            def setup(
                self,
                trainer,
                pl_module,
                stage: Optional[str] = None,
            ) -> None:
                self.device_stats_callback.setup(trainer, pl_module, stage)

            @rank_zero_only
            def on_train_batch_end(self, trainer, pl_module, *args) -> None:
                progress = 100 * (trainer.fit_loop.total_batch_idx + 1) / float(trainer.estimated_stepping_batches)
                if progress > 100:
                    self.work.progress = 100
                else:
                    self.work.progress = round(progress, 4)

                if not self.work.total_parameters:
                    if isinstance(trainer.strategy, DeepSpeedStrategy) and trainer.strategy.zero_stage_3:
                        total_parameters = sum(
                            deepspeed_param_size(p) if not _is_lazy_weight_tensor(p) else 0
                            for p in pl_module.parameters()
                            if p.requires_grad
                        )
                    else:
                        total_parameters = sum(p.numel() for p in pl_module.parameters() if p.requires_grad)
                    human_readable = get_human_readable_count(total_parameters)
                    self.work.total_parameters = str(human_readable)

                if trainer.checkpoint_callback.best_model_score:
                    self.best_model_score = float(trainer.checkpoint_callback.best_model_score)

                self.device_stats_callback.on_train_batch_end(trainer, pl_module, *args)

        def trainer_pre_fn(trainer, *args, **kwargs):
            callbacks = kwargs.get("callbacks", [])
            callbacks.append(ProgressCallback(self))
            kwargs["callbacks"] = callbacks
            return {}, args, kwargs

        tracer.add_traced(pl.Trainer, "__init__", pre_fn=trainer_pre_fn)

        return tracer


class ObjectiveLightningTrainingComponent(LightningTrainingComponent):
    def __init__(
        self,
        *args,
        experiment_id: int,
        experiment_name: str,
        logger: str,
        sweep_id: str,
        num_nodes: int = 1,
        **kwargs,
    ):
        super().__init__(
            *args,
            script_runner=PyTorchLightningObjective,
            logger=logger,
            sweep_id=sweep_id,
            experiment_id=experiment_id,
            experiment_name=experiment_name,
            num_nodes=num_nodes,
            **kwargs,
        )
        self.experiment_id = experiment_id
        self.experiment_name = experiment_name
        self.has_stopped = False
        self.pruned = False
        self.params = None
        self.restart_count = 0
        self.sweep_id = sweep_id
        self.reports = []
        self.has_stored = False

    def run(self, params: Optional[Dict[str, Any]] = None, restart_count: int = 0):
        self.params = params
        self.restart_count = restart_count
        super().run(params=params, restart_count=restart_count)

    @property
    def start_time(self):
        return self.ws[0].start_time

    @property
    def end_time(self):
        return self.ws[0].end_time

    @property
    def total_parameters(self):
        return self.ws[0].total_parameters

    @property
    def progress(self):
        return self.ws[0].progress

    @property
    def monitor(self):
        return self.ws[0].monitor

    @property
    def best_model_path(self):
        return self.ws[0].best_model_path

    @property
    def best_model_score(self):
        return self.ws[0].best_model_score

    @property
    def has_failed(self) -> bool:
        return any(w.has_failed for w in self.works())

    @property
    def status(self):
        return self.ws[0].status

    def stop(self):
        for w in self.works():
            w.stop()
        self.has_stopped = True

    @classmethod
    def distributions(cls):
        return {}



import urllib.parse
from typing import List

from lightning.app.storage import Drive, Mount
from lightning.app.structures import Dict

from lightning_hpo import Sweep
from lightning_hpo.commands.data.create import DataConfig
from lightning_hpo.commands.experiment.delete import DeleteExperimentCommand, DeleteExperimentConfig
from lightning_hpo.commands.experiment.run import RunExperimentCommand
from lightning_hpo.commands.experiment.show import ShowExperimentsCommand
from lightning_hpo.commands.experiment.stop import StopExperimentCommand, StopExperimentConfig
from lightning_hpo.commands.sweep.delete import DeleteSweepCommand, DeleteSweepConfig
from lightning_hpo.commands.sweep.run import RunSweepCommand, SweepConfig
from lightning_hpo.commands.sweep.show import ShowSweepsCommand
from lightning_hpo.commands.sweep.stop import StopSweepCommand, StopSweepConfig
from lightning_hpo.commands.tensorboard.stop import TensorboardConfig
from lightning_hpo.controllers.controller import Controller
from lightning_hpo.utilities.enum import Stage


class SweepController(Controller):

    model = SweepConfig

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tensorboard_sweep_id = None

    def on_reconcile_start(self, sweeps: List[SweepConfig]):
        # 1: Retrieve the tensorboard configs from the database
        if self.tensorboard_sweep_id is None:
            self.tensorboard_sweep_id = [c.sweep_id for c in self.db.select_all(TensorboardConfig)]

        for tensorboard in self.db.select_all(TensorboardConfig):
            work_name = urllib.parse.quote_plus(tensorboard.sweep_id)
            if work_name in self.r:
                self.r[work_name].logger_url = tensorboard.url

        # 2: Create the Sweeps
        for sweep in sweeps:
            id = sweep.sweep_id
            work_name = urllib.parse.quote_plus(id)
            if sweep.is_tensorboard():
                drive = Drive(f"lit://{id}")
                if id not in self.tensorboard_sweep_id:
                    self.tensorboard_sweep_id.append(id)
                    self.db.insert(TensorboardConfig(sweep_id=id, shared_folder=str(drive.drive_root)))
                elif sweep.stage in (Stage.FAILED, Stage.SUCCEEDED):
                    for tensorboard in self.db.select_all(TensorboardConfig):
                        if tensorboard.sweep_id == id:
                            tensorboard.desired_stage = Stage.STOPPED
                            self.db.update(tensorboard)

            if work_name not in self.r and sweep.stage != Stage.SUCCEEDED:
                all_data: List[DataConfig] = self.db.select_all(DataConfig)
                self.r[work_name] = Sweep.from_config(
                    sweep,
                    code={"drive": self.drive, "name": id},
                    mounts=[
                        Mount(data.source, mount_path=sweep.data[data.name] or data.mount_path)
                        for data in all_data
                        if data.name in sweep.data
                    ],
                )

    def on_reconcile_end(self, updates: List[SweepConfig]):
        for update in updates:
            if update.stage == Stage.SUCCEEDED:
                work_name = urllib.parse.quote_plus(update.sweep_id)
                for w in self.r[work_name].works():
                    w.stop()
                self.r.pop(work_name)

    def run_sweep(self, config: SweepConfig) -> str:
        work_name = urllib.parse.quote_plus(config.sweep_id)
        data_names = [data.name for data in self.db.select_all(DataConfig)]

        for data in config.data:
            if data not in data_names:
                return f"The provided Data '{data}' doesn't exist."

        if work_name not in self.r:
            self.db.insert(config)
            return f"Launched a Sweep '{config.sweep_id}'."
        return f"The current Sweep '{config.sweep_id}' is running. It couldn't be updated."

    def show_sweeps(self) -> List[Dict]:
        if self.db_url:
            return [sweep.dict() for sweep in self.db.select_all()]
        return []

    def stop_sweep(self, config: StopSweepConfig):
        work_name = urllib.parse.quote_plus(config.sweep_id)
        if work_name in self.r:
            # TODO: Add support for __del__ in lightning
            sweep: Sweep = self.r[work_name]
            for w in sweep.works():
                w.stop()
            sweep.stage = Stage.STOPPED
            sweep_config: SweepConfig = sweep.collect_model()
            for experiment in sweep_config.experiments.values():
                if experiment.stage == Stage.RUNNING:
                    experiment.stage = Stage.STOPPED
            self.db.update(sweep_config)
            return f"Stopped the sweep `{config.sweep_id}`"
        return f"We didn't find the sweep `{config.sweep_id}`"

    def delete_sweep(self, config: DeleteSweepConfig):
        sweeps: List[SweepConfig] = self.db.select_all()
        for sweep in sweeps:
            if config.name != sweep.sweep_id:
                continue
            else:
                if config.name in self.r:
                    sweep: Sweep = self.r[config.name]
                    for w in sweep.works():
                        w.stop()
                    sweep = sweep.collect_model()
                    del self.r[config.name]
                self.db.delete(sweep)
                return f"Deleted the sweep `{config.name}`"
        return f"We didn't find the sweep `{config.name}`"

    def delete_experiment(self, config: DeleteExperimentConfig):
        sweeps: List[SweepConfig] = self.db.select_all()
        for sweep in sweeps:
            if config.name != sweep.sweep_id:
                continue
            else:
                if config.name in self.r:
                    sweep: Sweep = self.r[config.name]
                    for w in sweep.works():
                        w.stop()
                    sweep = sweep.collect_model()
                    del self.r[config.name]
                self.db.delete(sweep)
                return f"Deleted the experiment `{config.name}`"
        return f"We didn't find the experiment `{config.name}`"

    def run_experiment(self, config: SweepConfig) -> str:
        work_name = urllib.parse.quote_plus(config.sweep_id)
        if work_name not in self.r:
            self.db.insert(config)
            return f"Launched an experiment '{config.sweep_id}'."
        return f"The experiment '{config.sweep_id}' is running. It can't be updated."

    def stop_experiment(self, config: StopExperimentConfig):
        sweeps_config: List[SweepConfig] = self.db.select_all()
        for sweep in sweeps_config:
            for experiment_id, experiment in enumerate(sweep.experiments.values()):
                if config.name == experiment.name:
                    if experiment.stage == Stage.SUCCEEDED:
                        return f"The current experiment `{experiment.name}` has already succeeded."
                    self.r[sweep.sweep_id].stop_experiment(experiment_id)
                    return f"The current experiment `{experiment.name}` has been stopped."
        return f"The current experiment `{config.name}` doesn't exist."

    def configure_commands(self):
        return [
            {"delete sweep": DeleteSweepCommand(self.delete_sweep)},
            {"run sweep": RunSweepCommand(self.run_sweep)},
            {"show sweeps": ShowSweepsCommand(self.show_sweeps)},
            {"stop sweep": StopSweepCommand(self.stop_sweep)},
            {"run experiment": RunExperimentCommand(self.run_experiment)},
            {"stop experiment": StopExperimentCommand(self.stop_experiment)},
            {"show experiments": ShowExperimentsCommand(self.show_sweeps)},
            {"delete experiment": DeleteExperimentCommand(self.delete_experiment)},
        ]
